#include "Game.h"

u8 running_ = FALSE;
u8 xScroll_ = 0;
u8 yScroll_ = 0;

void initGame(u8 level, u8 x, u8 y, u8 xScroll, u8 yScroll)
{
	u8 lightCurrentlyEquipped = 0;

    running_ = TRUE;

    // Initialise scrolling values to 0.
    xScroll_ = xScroll;
    yScroll_ = yScroll;	

	// Reset level sesion state before starting the level.
	resetLevelSessionState();

    //setup keys struct isDown values for user input
    initKeys();

    //setup the map
    initMap(level);

	// Setup the health state for enemies.
	initEnemies();

	// Setup boss state information.
	initBoss();

    // Clear level inventory values before starting the level.
	initLevelInventory();

	// Restore inventory values following a new level.
	restoreInventory();

    //setup initial player position and sizes
    initPlayer(x, y);

    //setup hud
	initHud();

	// Setup light sources.
	initLighting();
	
	lightCurrentlyEquipped = lightEquipped();

	// Register the light source if it hasn't previously been registered.
	if (lightCurrentlyEquipped && isLightRegistered(lightCurrentlyEquipped) == FALSE)
	{
		registerLightSource(lightCurrentlyEquipped, player.object.position.x, player.object.position.y, player.object.direction);
		
		// Activate the light following registration.
		activateLightByType(lightCurrentlyEquipped, TRUE);
	}

    // Clear Screen
    cpct_memset(CPCT_VMEM_START, 0, 0x4000);
}

void runGame()
{
    while(running_)
    {
        //read input from the player/user
        readInput();

        //run the game logic which will effect the display
        gameLogic();

		//wait for vsync
		cpct_waitVSYNC();

        // Draw the game display on the screen if the game is still running.
		// This extra if test is to prevent double draw for gesture attack.
		if (running_)
		{
			drawGame();
		}        
    }
}

void gameLogic()
{
    // Determine where the player can go next based on direction.
    determinePlayerNextMove(xScroll_, yScroll_);

    //move the player character
    movePlayer();

    // Run the logic on the player character (if it is their turn).
	logicPlayer(&xScroll_, &yScroll_);

	// Update lighting after player has moved.
	runLighting();

	// Run the logic for the projectile (if it is their turn).
	logicProjectile(xScroll_, yScroll_);

	// Run the logic for the enemies (if it is their turn).
	logicEnemies();

	// Run the logic for the boss (if it is their turn).
	logicBoss();	

    // Check if the player has ended playing the level (either in victory or defeat).
	if (levelEnded)
	{
		// Set running to false to transition to the next level or the end of the game.
		running_ = FALSE;
	}
}

void drawGame()
{
    //draw the player on the screen
    drawPlayer(xScroll_, yScroll_);

    //draw the content of the map on screen
    drawMap(xScroll_, yScroll_);

	// Draw the projectile on the screen.
	drawProjectile(xScroll_, yScroll_);

	// Draw the map enemies on the screen if visible.
	drawEnemies(xScroll_, yScroll_);

	// Draw the boss on the map if it has one.
	drawBossCells(xScroll_, yScroll_);

    // Draw an indicator on the screen detailing where the next player move will be.
    drawPlayerNextMoveIndicator(xScroll_, yScroll_);

    //hud will only update if hudRedraw is true
    drawHud();
}

void readInput()
{
	// Reads the standard input controls.
	readGameControls();

    if (cpct_isKeyPressed(gameControls.inventory))
	{
		u16 position = 0;
		u8 cellValue = 0;
		
		// Deactivate any currently equipped items before opening inventory screen.
		u8 lightCurrentlyEquipped = lightEquipped();
		activateLightByType(lightCurrentlyEquipped, FALSE);
		
        // Display the player inventory in its own dialog.
		inventoryDialog();

		// Light currently equipped needs to be updated following inventory changes.
		lightCurrentlyEquipped = lightEquipped();

		// Register the light source if it hasn't previously been registered.
		if (isLightRegistered(lightCurrentlyEquipped) == FALSE)
		{
			registerLightSource(lightCurrentlyEquipped, player.object.position.x, player.object.position.y, player.object.direction);
		}

		// Adding the item being dropped to the map at the current position of the player.
		if (inventoryToDropItem)
		{
			// Retrieve the current position of the player.
			position = XY_TO_POSITION((player.object.position.x), (player.object.position.y));

			// Combine inventory item with cell flags that we want.
			cellValue = inventoryToDropItem | CELL_ITEM_FLAG;

			// Oil lamp needs to have the light set.
			if (inventoryToDropItem == ITEM_OIL_LAMP)
			{
				cellValue |= CELL_LIGHT_FLAG;
			}

			// If the cell is currently visible it should remain visible
			if (mapCells[position] & CELL_LIGHT_FLAG)
			{
				cellValue |= CELL_LIGHT_FLAG;
			}

			// Add the item to the map at position.
			mapCells[position] = cellValue;
			
			// Remove the item from the inventory flags.
			removeItemFromInventory(inventoryToDropItem);

			// Update the position of the light source.
			updateLightPosition(inventoryToDropItem, player.object.position.x, player.object.position.y, player.object.direction);

			// Set light source to active.
			// TODO clarify behaviour about preactivating a lightsource before dropping it.
			activateLightByType(inventoryToDropItem, TRUE);

			// Reset inventory to drop item to 0 so it can only be dropped once.
			inventoryToDropItem = 0;
		}

		// Update position for light being equipped.
		updateLightPosition(lightCurrentlyEquipped, player.object.position.x, player.object.position.y, player.object.direction);

		// Activating the equipped light source if it has enough fuel.
		activateLightByType(lightCurrentlyEquipped, TRUE);

		// Redraw the hud with any of the equipped items.
		hudRedraw = TRUE;

		// Force a map redraw to cleanup after the dialog is closed.
		mapRedraw = TRUE;

		// Ensures that the exit dialog is not shown after exiting from the inventory dialog.
		gameControls.exitIsDown = FALSE;
	}

    if(gameControls.exitIsDown)
    {
        if(exitDialog())
        {
            running_ = FALSE;

			// Only exit the game completely on the entrance cave level.
			// This allows the player to quit back to the entrance cave but not the whole game.
			if (currentLevel == ENTRANCE_CAVE_LEVEL)
			{
				playerExiting = TRUE;
			}

			// Restores previously equiped items.
			restoreEquipedItems();

			// Clear all light sources so new ones can be initialised on level start.
			clearLightSources();
        }
        else
        {
            // Force a map redraw to cleanup after the dialog is closed.
		    mapRedraw = TRUE;
        }

		gameControls.exitIsDown = FALSE;
    }
}

void runGameSession()
{
	u8 startX = 0;
	u8 startY = 0;

	// Reset game state values for a new game.
	resetGameSessionState();

	// Setup the player inventory with zero items.
	initInventory();

    while(currentLevel != GAME_COMPLETE && !playerExiting)
    {
		startX = cave_start_position[currentLevel][0];
		startY = cave_start_position[currentLevel][1];

		// Default move indicator is white.
		G_move_indicator = G_highlighted_mask;

		// Reset level state.
		running_ = TRUE;

		// Defaults to dirt floors.
		G_tile_floor_sprite = G_tile_dirt_floor;

		if (currentLevel == ENTRANCE_CAVE_LEVEL)
		{
            // Set water level to true to animate in the interrupt handler.
            waterLevel = TRUE;

			// Set timer to > 320 so that interrupt handler will change it to water palette.
			timerCount = 500;

			// Water level has water floor (same as dirt but for different palette).
			G_tile_floor_sprite = G_tile_dirt_floor_river;

			// The player needs to be in a non scrolled part of the screen to determine how far the screen is scrolled.
			player.object.position.x = 1;
			player.object.position.y = 1;

			// Reset scrolling values before doing a catchup.
			xScroll_ = 0;
			yScroll_ = 0;

			catchupScroll(&xScroll_, &yScroll_, &preLevelTransitionPositionX, &preLevelTransitionPositionY);

			// Entrance cave starts either at 1,1 or at the position before a cave entrance is entered.
			initGame(currentLevel, preLevelTransitionPositionX, preLevelTransitionPositionY, xScroll_, yScroll_);

			// Update the cave graphics for the complete levels.
			updateCompleteEntranceCaves();
		}
		else
		{
			// The player needs to be in a non scrolled part of the screen to determine how far the screen is to be scrolled.
			player.object.position.x = 1;
			player.object.position.y = 1;

			// Reset scrolling values before doing a catchup.
			xScroll_ = 0;
			yScroll_ = 0;

			// Catch up the scrolling to startX, startY.
			catchupScroll(&xScroll_, &yScroll_, &startX, &startY);
			
			// Rivers level has animated water.
			if (currentLevel == RIVERS_LEVEL)
			{
				waterLevel = TRUE;

				// Set timer to > 320 so that interrupt handler will change it to water palette.
				timerCount = 500;

				// Water level has water floor (same as dirt but for different palette).
				G_tile_floor_sprite = G_tile_dirt_floor_river;
			}
			else
			{
				// Turn off animate in the interrupt handler.
				waterLevel = FALSE;

				// Reset the palette to the default.
				choosePalette(STANDARD_PALETTE);
			}
            
			// The plant based levels have a more green based palette.
			if (currentLevel == GROWING_PLANT_LEVEL || currentLevel == GESTURE_LEVEL)
			{
				// Plant level has plant floor.
				G_tile_floor_sprite = G_tile_plant_floor;

				// Using black indicator provides better contrast with the floor.
				G_move_indicator = G_highlighted_mask_black;

				// Change the palette with grass green as palette index 0.
				choosePalette(GRASS_PALETTE);
			}

			if (currentLevel == IMPASSABLE_LEVEL)
			{
				// On the impassable map the default cell is CELL_INTERACTABLE_FLAG | INTERACTABLE_DIRT_CRACK = 93.
				postMoveCell = 93;

				// Change the palette with pink as palette index 0.
				choosePalette(IMPASSABLE_PALETTE);
			}
			else
			{
				// Default square on move is the blank cell.
				postMoveCell = 0;
			}

			// Setup the game at position startx,starty for a normal level.
			initGame(currentLevel, startX, startY, xScroll_, yScroll_);
		}

        //start the game for the current level
        runGame();

		// Move to the next level based on the player position.
		currentLevel = chooseNextCave(player.object.position.x, player.object.position.y);

		// Change all caves to a particular level for testing.
		//currentLevel = 10;
		
		levelEnded = LEVEL_READY;
		// TODO update game state if level was won.
    }
}

u8 chooseNextCave(u8 x, u8 y)
{
	// Only choose custom level from the entrance cave level.
	if (currentLevel == ENTRANCE_CAVE_LEVEL)
	{
		u8 i = 0;

		for (i = 0; i < MAP_ENTRANCECAVE_ENTRANCE_ELEMENT_COUNT; i += 2)
		{
			// Match the player position against the cave entrances in the entrance cave map.
			if (cave_entrances[0][i] == x && cave_entrances[0][i + 1] == y)
			{
				// The order of entrances determines which map is loaded.
				// +1 as the entrance cave is level 0.
				return (i / 2) + 1;
			}
		}
	}
	else if (currentLevel == GAME_COMPLETE)
	{
		// Game complete.
		return GAME_COMPLETE;
	}
	
	// Return to the entrance cave level.
	return 0;
}

void updateCompleteEntranceCaves()
{
	u8 i = 0;
	u8 allComplete = TRUE;
	u8 finalX = 255;
	u8 finalY = 255;
	
	// Updates the caves in the entrance cave to complete if the cave has been completed.
	for (i = 0; i < CAVE_COUNT * 2; i += 2)
	{		
		u8 x = cave_entrances[ENTRANCE_CAVE_LEVEL][i];
		u8 y = cave_entrances[ENTRANCE_CAVE_LEVEL][i + 1];

		if (caveComplete[i / 2])
		{
			// CELL_INTERACTABLE_FLAG | INTERACTABLE_CAVE_ENTRANCE_COMPLETED = 64 | 9 | = 73
			MAPA(x, y) = 73;
		}
		else
		{
			if ((MAPA(x, y) & CELL_GFX_RANGE) == INTERACTABLE_CAVE_ENTRANCE_LOCKED)
			{
				// Save the location of the boss level.
				finalX = x;
				finalY = y;
			}
			else
			{
				allComplete = FALSE;
			}
		}
	}
	
	// Provides access to the last level once all the caves are complete.
	if (allComplete)
	{
		MAPA(finalX, finalY) = CELL_INTERACTABLE_FLAG | INTERACTABLE_CAVE_ENTRANCE_OPEN;
	}
}