#include "Lighting.h"

u8 lightSourceCount;
LightSource lightSources[MAX_LIGHT_SOURCE_COUNT];
u8 lightingChange = TRUE;

void initLighting()
{
	u8 i = 0;
	lightingChange = TRUE;

	lightSourceCount = 0;

	for (i = 0; i < mapStaticLightSourceCount; i += 2)
	{
		// Registers the light in the system.
		registerLightSource(STATIC_LIGHT_SOURCE_TYPE, mapStaticLightSources[i], mapStaticLightSources[i+1], DIRECTION_UP);
		
		// Sets the light source to active.
		lightSources[i/2].active = TRUE;
	}
}

void activateLightByPosition(u8 x, u8 y, u8 active)
{
    u8 i = 0;

    for (i = 0; i < lightSourceCount; ++i)
    {
        if(lightSources[i].position.x == x && lightSources[i].position.y == y)
        {
			lightSources[i].active = active;
        }
    }

	lightingChange = TRUE;
}

void activateLightByType(u8 type, u8 active)
{
	u8 i = 0;

	for (i = 0; i < lightSourceCount; ++i)
	{
		if (lightSources[i].type == type)
		{
			lightSources[i].active = active;
		}
	}

	lightingChange = TRUE;
}

void updateLightPosition(u8 type, u8 x, u8 y, u8 direction)
{
    // We are relying on type being unique for moving lights.
    u8 i = 0;

    for (i = 0; i < lightSourceCount; ++i)
    {
        if(lightSources[i].type == type)
        {
            // Save current position to previous position.
            lightSources[i].previousPosition.x = lightSources[i].position.x;
            lightSources[i].previousPosition.y = lightSources[i].position.y;
            lightSources[i].previousDirection = lightSources[i].direction;

            // Update current position.
            lightSources[i].position.x = x;
            lightSources[i].position.y = y;
            lightSources[i].direction = direction;
        }
    }

	lightingChange = TRUE;
}

void registerLightSource(u8 type, u8 x, u8 y, u8 direction)
{
	// Make sure there is a valid light source before attempting to register it.
	if (type)
	{
		// Add the light source to the array.
		lightSources[lightSourceCount].previousPosition.x = x;
		lightSources[lightSourceCount].previousPosition.y = y;
		lightSources[lightSourceCount].previousDirection = direction;
		lightSources[lightSourceCount].position.x = x;
		lightSources[lightSourceCount].position.y = y;
		lightSources[lightSourceCount].direction = direction;
		lightSources[lightSourceCount].type = type;

		// It can be activated in the activate function.
		lightSources[lightSourceCount].active = FALSE;

		// Increase the count.
		++lightSourceCount;
	}	
}

void clearLightSources()
{
	lightSourceCount = 0;
}

void runLighting()
{
	// Only draw lighting on a map or lighting change.
	if (lightingChange)
	{
		u8 i = 0;

		// Clears the light flags for all the light sources at all possible positions.
		for (i = 0; i < lightSourceCount; ++i)
		{
			LightSource* lightSource = &lightSources[i];

			// Deactivate the light if it was at some previous position and has moved.
			displayLight(lightSource->previousPosition.x, lightSource->previousPosition.y, lightSource->type, FALSE, lightSource->previousDirection);

			// Dactivate the light at the current position for if it hasn't moved but still has a previous position that is different.
			displayLight(lightSource->position.x, lightSource->position.y, lightSource->type, FALSE, lightSource->direction);
		}

		// Sets the light flags for all the active light sources at their current positions.
		for (i = 0; i < lightSourceCount; ++i)
		{
			LightSource* lightSource = &lightSources[i];

			// Only displays the active light sources.
			if (lightSource->active)
			{
				// Activate the light the light for where it is currently located.
				displayLight(lightSource->position.x, lightSource->position.y, lightSource->type, TRUE, lightSource->direction);
			}
		}

		lightingChange = FALSE;
	}	
}

u8 isLightRegistered(u8 type)
{
	u8 i = 0;
	
	for (i = 0; i < lightSourceCount; ++i)
	{
		if (lightSources[i].type == type)
		{
			return TRUE;
		}
	}

	return FALSE;
}

void displayLight(u8 playerX, u8 playerY, u8 activeLightType, u8 lightOn, u8 direction)
{
	if (activeLightType == ITEM_OIL_LAMP || activeLightType == STATIC_LIGHT_SOURCE_TYPE)
	{
		//7,5,3,1 from x,y
		lightRightAlternate(playerX, playerY, lightCellsOutY, lightOn);

		//1,3,5,7 from x,y starting at 7
		lightLeftAlternate(playerX, playerY, lightCellsOutY, lightOn);
	}
	else if (activeLightType == ITEM_ELECTRIC_TORCH)
	{
		if (direction == DIRECTION_UP)
		{
			//3x4 beam from x,y
			lightNorth(playerX, playerY, lightBeamLength, lightOn);
		}
		else if (direction == DIRECTION_DOWN)
		{
			lightSouth(playerX, playerY, lightBeamLength, lightOn);
		}
		else if (direction == DIRECTION_LEFT)
		{
			lightBeamWest(playerX, playerY, lightBeamLength, 3, lightOn);
		}
		else
		{
			lightBeamEast(playerX, playerY, lightBeamLength, 3, lightOn);
		}
	}
	else if (activeLightType == ITEM_SCUBA)
	{
		u8 cell = MAPA(playerX, playerY) & CELL_GFX_RANGE;

		// Scuba gear only works if the player is in the water.
		if (cell == TILE_WATER_HORIZONTAL || cell == TILE_WATER_VERTICAL)
		{
			if (direction == DIRECTION_UP)
			{
				//3x4 beam from x,y
				lightNorth(playerX, playerY, lightBeamLength, lightOn);
			}
			else if (direction == DIRECTION_DOWN)
			{
				lightSouth(playerX, playerY, lightBeamLength, lightOn);
			}
			else if (direction == DIRECTION_LEFT)
			{
				lightBeamWest(playerX, playerY, lightBeamLength, 3, lightOn);
			}
			else
			{
				lightBeamEast(playerX, playerY, lightBeamLength, 3, lightOn);
			}
		}
	}
}

void lightRightAlternate(u8 x, u8 y, u8 cellsOutY, u8 lightOn)
{
	// Setting the light flags to the right.
	i8 i = 0;
	i8 j = 0;
	u16 position = XY_TO_POSITION(x, y);

	for (i = cellsOutY; i >= 0; --i)
	{
		// We are adding j to the position using a range from negative i to i.
		// position - i -> position + i
		for (j = i; j >= -i; --j)
		{
			// The first part is the top of the screen, the second part is for the bottom of the screen.
			if (y + j >= 0 && y + j < MAP_TILES_Y)
			{
				if (lightOn)
				{
					// If a cell is not an interactable or an item and it has a graphic then it must be a solid wall.
					// TODO handle water tiles which you are allowed to see.
					if (!(mapCells[position + j] & CELL_INTERACTABLE_OR_ITEM_FLAG) && (mapCells[position + j] && mapCells[position + j] < CELL_LIGHT_BLOCKER_GFX_RANGE))
					{
						// Do not set the flag as it is a solid wall.
					}
					else
					{
						// Test that j != i to make sure this isn't the first column.
						// Test that the earlier cell (x asis) is not dark.

						if (i != cellsOutY && !(mapCells[position + j + MAP_TILES_Y] & CELL_LIGHT_FLAG))
						{
							// In this instance we don't set the flag before earlier column cell is not visible.
							// This lets us block out the light on the x asix.
						}
						else
						{
							mapCells[position + j] |= CELL_LIGHT_FLAG;
						}
					}

					// We have reached the cell one above the origin (x,y+1)
					if (j == 1)
					{
						// Below x light blocker (vertical).
						u8 k = 0;

						// Replicate darkness from one above the origin (x,y+1) to cells below it (y+2 and on).
						for (k = 0; k < i - 1; ++k)
						{
							// Check if the first cell is dark.
							if (!(mapCells[position + j + k] & CELL_LIGHT_FLAG))
							{
								// If the cell above is dark then the cell below it must also be set to dark (remove the flag).
								mapCells[position + j + k + 1] &= ~CELL_LIGHT_FLAG;
							}
						}
					}
					else if (j < -1) // From (x, y-2).
					{
						// Above x light blocker (vertical).
						// We can test on the previous cell easily because the function progresses from bottom to top of the column.
						if (!(mapCells[position + j + 1] & CELL_LIGHT_FLAG))
						{
							// If the previous (above) cell is dark then remove the light flag from the current cell.
							mapCells[position + j] &= ~CELL_LIGHT_FLAG;
						}
					}
				}
				else
				{
					mapCells[position + j] &= ~CELL_LIGHT_FLAG;
				}
			}
		}

		// Move to the prev column.
		position -= MAP_TILES_Y;

		// Out of range so break out of the loop instead of getting memory corruption.
		if (position >= MAP_CELL_COUNT)
		{
			break;
		}
	}
}

void lightLeftAlternate(u8 x, u8 y, u8 cellsOutY, u8 lightOn)
{
	// Setting the light flags to the left.
	i8 i = 0;
	i8 j = 0;
	u16 position = XY_TO_POSITION(x, y);

	for (i = cellsOutY; i >= 0; --i)
	{
		// We are adding j to the position using a range from negative i to i.
		// position - i -> position + i
		for (j = i; j >= -i; --j)
		{
			// The first part is the top of the screen, the second part is for the bottom of the screen.
			if (y + j >= 0 && y + j < MAP_TILES_Y)
			{
				if (lightOn)
				{
					// If a cell is not an interactable or an item and it has a graphic then it must be a solid wall.
					// TODO handle water tiles which you are allowed to see.
					if (!(mapCells[position + j] & CELL_INTERACTABLE_OR_ITEM_FLAG) && (mapCells[position + j] && mapCells[position + j] < CELL_LIGHT_BLOCKER_GFX_RANGE))
					{
						// Do not set the flag as it is a solid wall.
					}
					else
					{
						// Test that j != i to make sure this isn't the first column.
						// Test that the earlier cell (x asis) is not dark
						if (i != cellsOutY && !(mapCells[position + j - MAP_TILES_Y] & CELL_LIGHT_FLAG))
						{
							// In this instance we don't set the flag before earlier column cell is not visible.
							// This lets us block out the light on the x asix.
						}
						else
						{
							mapCells[position + j] |= CELL_LIGHT_FLAG;
						}
					}

					// We have reached the cell one above the origin (x,y+1)
					if (j == 1)
					{
						// Below x light blocker (vertical).
						u8 k = 0;

						// Replicate darkness from one above the origin (x,y+1) to cells below it (y+2 and on).
						for (k = 0; k < i - 1; ++k)
						{
							// Check if the first cell is dark.
							if (!(mapCells[position + j + k] & CELL_LIGHT_FLAG))
							{
								// If the cell above is dark then the cell below it must also be set to dark (remove the flag).
								mapCells[position + j + k + 1] &= ~CELL_LIGHT_FLAG;
							}
						}
					}
					else if (j < -1) // From (x, y-2).
					{
						// Above x light blocker (vertical).
						// We can test on the previous cell easily because the function progresses from bottom to top of the column.
						if (!(mapCells[position + j + 1] & CELL_LIGHT_FLAG))
						{
							// If the previous (above) cell is dark then remove the light flag from the current cell.
							mapCells[position + j] &= ~CELL_LIGHT_FLAG;
						}
					}
				}
				else
				{
					mapCells[position + j] &= ~CELL_LIGHT_FLAG;
				}
			}
		}

		// Move to the next column.
		position += MAP_TILES_Y;

		// Out of range so break out of the loop instead of getting memory corruption.
		if (position >= MAP_CELL_COUNT)
		{
			break;
		}
	}
}

void lightNorth(u8 x, u8 y, u8 cellsOutY, u8 lightOn)
{
	u16 position = XY_TO_POSITION(x, y);
	u8 i = 0;
	i8 j = 0;
	u8 beamWidth = 3;

	for (i = 0; i < cellsOutY; ++i)
	{
		// We are adding j to the position using a range from negative i to i.
		// x - 1 -> x + 1 (where width is 3)
		for (j = -(beamWidth - 2); j <= beamWidth - 2; ++j)
		{
			if (lightOn)
			{
				// If a cell is not an interactable or an item and it has a graphic then it must be a solid wall.
				// TODO handle water tiles which you are allowed to see.
				if (!(mapCells[position + j * MAP_TILES_Y] & CELL_INTERACTABLE_OR_ITEM_FLAG) && (mapCells[position + j * MAP_TILES_Y] && mapCells[position + j * MAP_TILES_Y] < CELL_LIGHT_BLOCKER_GFX_RANGE))
				{
					// Do not set the flag as it is a solid wall.
				}
				else
				{
					if (i != 0 && !(mapCells[1 + position + j * MAP_TILES_Y] & CELL_LIGHT_FLAG))
					{
						// The first row is always shown.
						// If the previous cell hasn't been set to dark then the subsequent cell is not set to light.
					}
					else
					{
						mapCells[position + j * MAP_TILES_Y] |= CELL_LIGHT_FLAG;
					}
				}
			}
			else
			{
				mapCells[position + j * MAP_TILES_Y] &= ~CELL_LIGHT_FLAG;
			}
		}

		// Move one cell up (north) the grid.
		--position;

		// Out of range so break out of the loop instead of getting memory corruption.
		if (position >= MAP_CELL_COUNT)
		{
			break;
		}
	}
}

void lightSouth(u8 x, u8 y, u8 cellsOutY, u8 lightOn)
{
	u16 position = XY_TO_POSITION(x, y);
	u8 i = 0;
	i8 j = 0;
	u8 beamWidth = 3;

	for (i = 0; i < cellsOutY; ++i)
	{
		// We are adding j to the position using a range from negative i to i.
		// x - 1 -> x + 1 (where width is 3)
		for (j = -(beamWidth - 2); j <= beamWidth - 2; ++j)
		{
			if (lightOn)
			{
				// If a cell is not an interactable or an item and it has a graphic then it must be a solid wall.
				// TODO handle water tiles which you are allowed to see.
				if (!(mapCells[position - j * MAP_TILES_Y] & CELL_INTERACTABLE_OR_ITEM_FLAG) && (mapCells[position - j * MAP_TILES_Y] && mapCells[position - j * MAP_TILES_Y] < CELL_LIGHT_BLOCKER_GFX_RANGE))
				{
					// Do not set the flag as it is a solid wall.
				}
				else
				{
					if (i != 0 && !(mapCells[-1 + position - j * MAP_TILES_Y] & CELL_LIGHT_FLAG))
					{
						// The first row is always shown.
						// If the previous cell hasn't been set to dark then the subsequent cell is not set to light.
					}
					else
					{
						mapCells[position - j * MAP_TILES_Y] |= CELL_LIGHT_FLAG;
					}
				}
			}
			else
			{
				mapCells[position - j * MAP_TILES_Y] &= ~CELL_LIGHT_FLAG;
			}
		}

		// Move one cell down (south) the grid.
		++position;

		// Out of range so break out of the loop instead of getting memory corruption.
		if (position >= MAP_CELL_COUNT)
		{
			break;
		}
	}
}

void lightBeamEast(u8 x, u8 y, u8 beamLength, u8 beamWidth, u8 lightOn)
{
	// Setting the light flags to the right.
	u16 position = XY_TO_POSITION(x, y);
	u8 i = 0;
	i8 j = 0;

	// The length is how may cells the beam reaches across x.
	for (i = 0; i < beamLength; ++i)
	{
		// The width is the number of cells the beam reaches along y.
		// We are adding j to the position using a range from negative i to i.
		// position - i -> position + i
		// x - 1 -> x + 1 (where width is 3)
		for (j = -(beamWidth - 2); j <= beamWidth - 2; ++j)
		{
			// The first part is the top of the screen, the second part is for the bottom of the screen.
			if (y + j >= 0 && y + j < MAP_TILES_Y)
			{
				if (lightOn)
				{
					// If a cell is not an interactable or an item and it has a graphic then it must be a solid wall.
					// TODO handle water tiles which you are allowed to see.
					if (!(mapCells[position + j] & CELL_INTERACTABLE_OR_ITEM_FLAG) && (mapCells[position + j] && mapCells[position + j] < CELL_LIGHT_BLOCKER_GFX_RANGE))
					{
						// Do not set the flag as it is a solid wall.
					}
					else
					{
						if (i != 0 && !(mapCells[position + j - MAP_TILES_Y] & CELL_LIGHT_FLAG))
						{
							// In this instance we don't set the flag if the previous cell (left of current) is currently dark.
							// This lets us block out the light on the x asix.
						}
						else
						{
							mapCells[position + j] |= CELL_LIGHT_FLAG;
						}
					}
				}
				else
				{
					mapCells[position + j] &= ~CELL_LIGHT_FLAG;
				}
			}
		}

		// Move to the next column.
		position += MAP_TILES_Y;

		// Out of range so break out of the loop instead of getting memory corruption.
		if (position >= MAP_CELL_COUNT)
		{
			break;
		}
	}
}

void lightBeamWest(u8 x, u8 y, u8 beamLength, u8 beamWidth, u8 lightOn)
{
	// Setting the light flags to the right.
	u16 position = XY_TO_POSITION(x, y);
	u8 i = 0;
	i8 j = 0;

	// The length is how may cells the beam reaches across x.
	for (i = 0; i < beamLength; ++i)
	{
		// The width is the number of cells the beam reaches along y.
		// We are adding j to the position using a range from negative i to i.
		// position - i -> position + i
		// x - 1 -> x + 1 (where width is 3)
		for (j = -(beamWidth - 2); j <= beamWidth - 2; ++j)
		{
			// The first part is the top of the screen, the second part is for the bottom of the screen.
			if (y + j >= 0 && y + j < MAP_TILES_Y)
			{
				if (lightOn)
				{
					// If a cell is not an interactable or an item and it has a graphic then it must be a solid wall.
					// TODO handle water tiles which you are allowed to see.
					if (!(mapCells[position + j] & CELL_INTERACTABLE_OR_ITEM_FLAG) && (mapCells[position + j] && mapCells[position + j] < CELL_LIGHT_BLOCKER_GFX_RANGE))
					{
						// Do not set the flag as it is a solid wall.
					}
					else
					{
						if (i != 0 && !(mapCells[position + j + MAP_TILES_Y] & CELL_LIGHT_FLAG))
						{
							// In this instance we don't set the flag if the previous cell (right of current) is currently dark.
							// This lets us block out the light on the x asix.
						}
						else
						{
							mapCells[position + j] |= CELL_LIGHT_FLAG;
						}
					}
				}
				else
				{
					mapCells[position + j] &= ~CELL_LIGHT_FLAG;
				}
			}
		}

		// Move to the next column to the left.
		position -= MAP_TILES_Y;

		// Out of range so break out of the loop instead of getting memory corruption.
		if (position >= MAP_CELL_COUNT)
		{
			break;
		}
	}
}

u8 lightEquipped()
{
	if (inventoryLeftItem == ITEM_OIL_LAMP || inventoryRightItem == ITEM_OIL_LAMP)
	{
		if (levelOilFuel)
		{
			return ITEM_OIL_LAMP;
		}
	}
	else if (inventoryLeftItem == ITEM_ELECTRIC_TORCH || inventoryRightItem == ITEM_ELECTRIC_TORCH)
	{
		if (levelBatteryFuel)
		{
			return ITEM_ELECTRIC_TORCH;
		}
	}
	else if (inventoryLeftItem == ITEM_SCUBA || inventoryRightItem == ITEM_SCUBA)
	{
		return ITEM_SCUBA;
	}

	// The light does not get treated as equipped if there is no fuel.
	return 0;
}

u8 nearestLight(u8* x, u8* y)
{
	u8 i = 0;
	u16 sourceDistance = distance(*x, *y);
	u16 shortestDistance = 65535;
	u8 shortestDistanceIndex = 255;

	for (i = 0; i < lightSourceCount; ++i)
	{
		// Only active light sources are compared.
		if (lightSources[i].active)
		{
			u16 destinationDistance = distance(lightSources[i].position.x, lightSources[i].position.y);
			i16 computedDistance = absoluteValue(sourceDistance - destinationDistance);

			if (computedDistance < shortestDistance)
			{
				shortestDistanceIndex = i;
				shortestDistance = computedDistance;
			}
		}		
	}
	
	if (shortestDistanceIndex != 255)
	{
		// Returns the position of the nearest light source.
		*x = lightSources[shortestDistanceIndex].position.x;
		*y = lightSources[shortestDistanceIndex].position.y;
		return 1;
	}
	else
	{
		// No active light sources found.
		return 0;
	}
}
